import os


def suppress_stdout():
    global NULL_FDS
    NULL_FDS = [os.open(os.devnull, os.O_RDWR) for x in range(2)]
    global BACKUP_FDS
    BACKUP_FDS = os.dup(1), os.dup(2)
    os.dup2(NULL_FDS[0], 1)
    os.dup2(NULL_FDS[1], 2)


def restore_stdout():
    os.dup2(BACKUP_FDS[0], 1)
    os.dup2(BACKUP_FDS[1], 2)
    os.close(NULL_FDS[0])
    os.close(NULL_FDS[1])


#!/usr/bin/env python
import sys
import argparse
import uuid
from cwl_airflow.utils.mute import suppress_stdout, restore_stdout
suppress_stdout()
# Suppress output
from airflow.bin.cli import scheduler
from cwl_airflow.utils.func import export_job_file, update_args, update_config, export_dags, create_folders, get_demo_workflow
from cwl_airflow.utils.utils import get_workflow_output, normalize_args, exit_if_unsupported_feature
# Restore output
restore_stdout()


def arg_parser():
    parent_parser = argparse.ArgumentParser(add_help=False)
    general_parser = argparse.ArgumentParser(description='cwl-airflow')

    subparsers = general_parser.add_subparsers()
    subparsers.required = True

    init_parser = subparsers.add_parser('init', help="Init cwl-airflow", parents=[parent_parser])
    init_parser.set_defaults(func=run_init)
    init_parser.add_argument("-l", "--limit", dest='limit', type=int, help="Limit job concurrancy", default=10)
    init_parser.add_argument("-t", "--timeout", dest='dag_timeout', type=int, help="How long before timing out a python file import while filling the DagBag", default=30)
    init_parser.add_argument("-i", "--interval", dest='dag_interval', type=int, help="After how much time a new DAGs should be picked up from the filesystem", default=0)
    init_parser.add_argument("-r", "--refresh", dest='web_interval', type=int, help="Webserver refresh interval", default=30)
    init_parser.add_argument("-w", "--workers", dest='web_workers', type=int, help="Webserver workers refresh batch size", default=1)
    init_parser.add_argument("-p", "--threads", dest='threads', type=int, help="Max scheduler threads", default=2)

    run_parser = subparsers.add_parser('run', help="Run workflow", parents=[parent_parser])
    run_parser.set_defaults(func=run_job)
    run_parser.add_argument("-o", "--outdir", dest='output_folder', type=str, help="Output directory, default current directory", default=".")
    run_parser.add_argument("-t", "--tmp", dest='tmp_folder', type=str, help="Folder to store temporary data")
    run_parser.add_argument("-u", "--uid", dest='uid', type=str, help="Unique ID", default=str(uuid.uuid4()))
    run_parser.add_argument("workflow", type=str)
    run_parser.add_argument("job", type=str)

    demo_parser = subparsers.add_parser('demo', help="Run demo workflows", parents=[parent_parser])
    demo_parser.set_defaults(func=run_demo)
    demo_parser.add_argument("-o", "--outdir", dest='output_folder', type=str, help="Output directory, default current directory", default=".")
    demo_parser.add_argument("-t", "--tmp", dest='tmp_folder', type=str, help="Folder to store temporary data")
    demo_parser.add_argument("-u", "--uid", dest='uid', type=str, help="Unique ID", default=str(uuid.uuid4()))
    demo_parser.add_argument("workflow", type=str)

    return general_parser


def run_demo(args):
    if not args.workflow:
        print("Available workflows to run:")
        for wf in get_demo_workflow():
            print("-",wf["workflow"]["name"])
    else:
        selected_demo = get_demo_workflow(args.workflow)
        args.workflow = selected_demo[0]["workflow"]["path"]
        args.job = selected_demo[0]["job"]["path"]
        run_job(args)

def run_init(args):
    update_config(args)
    create_folders()
    export_dags()


def run_job(args):
    suppress_stdout()
    exit_if_unsupported_feature(args.workflow)
    export_job_file(args)
    update_args(args)
    scheduler(args)
    restore_stdout()
    print(get_workflow_output(args.dag_id))


def main(argsl=None):
    if argsl is None:
        argsl = sys.argv[1:]
    argsl.append("")  # To avoid raising error when argsl is empty
    args, _ = arg_parser().parse_known_args(argsl)
    args = normalize_args(args, skip_list=["func", "uid", "limit", "dag_timeout", "dag_interval", "threads", "web_interval", "web_workers"])
    args.func(args)


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



import os
from json import dumps
import configparser
from datetime import datetime
from airflow import conf as conf
from airflow.settings import DAGS_FOLDER, AIRFLOW_HOME
from airflow.bin.cli import get_dag, CLIFactory
from cwl_airflow.utils.utils import (set_logger,
                                     gen_dag_id,
                                     get_folder,
                                     load_job,
                                     list_files,
                                     export_to_file,
                                     norm_path,
                                     get_files)
from cwl_airflow.dag_components.cwldag import CWLDAG
from cwl_airflow.dag_components.jobdispatcher import JobDispatcher
from cwl_airflow.dag_components.jobcleanup import JobCleanup


def get_demo_workflow(target_wf=None, job_ext=".json"):
    workflows = get_files(norm_path(os.path.join(os.path.dirname(os.path.abspath(os.path.join(__file__, "../"))), "tests/cwl")))
    jobs = get_files(norm_path(os.path.join(os.path.dirname(os.path.abspath(os.path.join(__file__, "../"))), "tests/job")))
    combined_data = []
    for wf_name, wf_path in workflows.items():
        job_name = os.path.splitext(wf_name)[0] + job_ext
        if job_name in jobs:
            combined_data.append({"workflow": {"name": wf_name,
                                               "path": wf_path},
                                  "job": {"name": job_name,
                                          "path": jobs[job_name]}
                                  })
    return [item for item in combined_data if item["workflow"]["name"] == os.path.basename(target_wf)] if target_wf else combined_data


def export_job_file(args):
    job_entry = load_job(args.job)
    del job_entry["id"]
    job_entry['workflow'] = job_entry.get("workflow", args.workflow)
    job_entry['output_folder'] = job_entry.get("output_folder", args.output_folder)
    job_entry["uid"] = job_entry.get("uid", args.uid)
    job_entry['tmp_folder'] = job_entry.get("tmp_folder", args.tmp_folder)
    root, ext = os.path.splitext(os.path.basename(args.job))
    args.job = os.path.join(conf.get('cwl', 'jobs'), root + "-" + job_entry["uid"] + ext)
    export_to_file(args.job, dumps(job_entry, indent=4))


def update_args(args):
    vars(args).update({arg_name: arg_value.default for arg_name, arg_value in CLIFactory.args.items()
                       if arg_name in CLIFactory.subparsers_dict['scheduler']['args']})
    args.dag_id = gen_dag_id(os.path.join(conf.get('cwl', 'jobs'), os.path.basename(args.job)))
    args.num_runs = len(get_dag(args).tasks) + 3


def get_active_jobs(jobs_folder, limit=10):
    """
    :param jobs_folder: job_folder: abs path to the folder with job json files  
    :param limit: max number of jobs to return
    :return: 
    """
    all_jobs = []
    for job_path in list_files(abs_path=jobs_folder, ext=[".json", ".yml", ".yaml"]):
        job_content = load_job(job_path)
        all_jobs.append({"content": job_content,
                         "path": job_path,
                         "creation_date": datetime.fromtimestamp(os.path.getctime(job_path)),
                         "dag_id": gen_dag_id(job_path)})
    all_jobs = sorted(all_jobs, key=lambda k: k["creation_date"], reverse=True)
    return all_jobs[:limit]


def make_dag(job):
    """
    :param job: {"content": job_entry,
                 "path": job,
                 "creation_date": datetime.fromtimestamp(os.path.getctime(job_path)),
                 "dag_id": gen_dag_id(job_entry["workflow"], job_path)}
    :return:
    """
    set_logger()
    default_args = {
        'start_date': job["creation_date"],
        "job_data":   job
    }

    dag = CWLDAG(
        dag_id=job["dag_id"],
        schedule_interval='@once',
        default_args=default_args)
    dag.create()
    dag.assign_job_dispatcher(JobDispatcher(dag=dag))
    dag.assign_job_cleanup(JobCleanup(dag=dag))
    return dag


def update_config(args):
    with open(conf.AIRFLOW_CONFIG, 'w') as output_stream:
        try:
            conf.conf.add_section('cwl')
        except configparser.DuplicateSectionError:
            pass
        conf.set('core', 'dags_are_paused_at_creation', 'False')
        conf.set('core', 'load_examples', 'False')
        conf.set('cwl', 'jobs', os.path.join(AIRFLOW_HOME, 'jobs'))
        conf.set('cwl', 'limit', str(args.limit))
        conf.set('core', 'dagbag_import_timeout', str(args.dag_timeout))
        conf.set('scheduler', 'min_file_process_interval', str(args.dag_interval))
        conf.set('scheduler', 'max_threads', str(args.threads))
        conf.set('webserver', 'worker_refresh_interval', str(args.web_interval))
        conf.set('webserver', 'worker_refresh_batch_size', str(args.web_workers))
        conf.conf.write(output_stream)


def export_dags():
    dag_content = u"#!/usr/bin/env python3\nfrom airflow import DAG\nfrom cwl_airflow.create_dag import create_dags\nfor id, dag in create_dags().items():\n    globals()[id] = dag"
    export_to_file(os.path.join(DAGS_FOLDER, "cwl_dag.py"), dag_content)


def create_folders():
    get_folder(conf.get('cwl', 'jobs'))
    get_folder(DAGS_FOLDER)
